@@ -83,7 +83,7 @@ mm_real *
 mm_real_new (MMRealFormat format, MMRealSymm symm, const int m, const int n, const int nnz)
 {
 	mm_real	*x;
-	bool		symmetric;
+	bool	symmetric;
 
 	if (!is_format_valid (format))
 		error_and_exit ("mm_real_new", "invalid MMRealFormat format.", __FILE__, __LINE__);
@@ -106,10 +106,12 @@ mm_real_new (MMRealFormat format, MMRealSymm symm, const int m, const int n, con
 		x->i = (int *) malloc (x->nnz * sizeof (int));
 		x->p = (int *) malloc ((x->n + 1) * sizeof (int));
 		if (x->i == NULL || x->p == NULL) error_and_exit ("mm_real_new", "cannot allocate memory.", __FILE__, __LINE__);
+		// initialize x->p[0]
+		x->p[0] = 0;
 	} else mm_set_array (&x->typecode);
 
 	x->symm = symm;
-	// typecode[3] = 'S'
+	// typecode[3] = 'G' -> 'S'
 	if (symmetric) mm_set_symmetric (&x->typecode);
 
 	if (!is_type_supported (x->typecode)) {
@@ -117,18 +119,10 @@ mm_real_new (MMRealFormat format, MMRealSymm symm, const int m, const int n, con
 		sprintf (msg, "matrix type does not supported :[%s].", mm_typecode_to_str (x->typecode));
 		error_and_exit ("mm_real_new", msg, __FILE__, __LINE__);
 	}
-	if (format == MM_REAL_SPARSE) mm_set_coordinate (&x->typecode);
 
 	// allocate arrays
 	x->data = (double *) malloc (x->nnz * sizeof (double));
 	if (x->data == NULL) error_and_exit ("mm_real_new", "cannot allocate memory.", __FILE__, __LINE__);
-	if (mm_real_is_sparse (x)) {
-		x->i = (int *) malloc (x->nnz * sizeof (int));
-		x->p = (int *) malloc ((x->n + 1) * sizeof (int));
-		if (x->i == NULL || x->p == NULL) error_and_exit ("mm_real_new", "cannot allocate memory.", __FILE__, __LINE__);
-		// initialize x->p[0]
-		x->p[0] = 0;
-	}
 
 	return x;
 }
@@ -138,10 +132,8 @@ void
 mm_real_free (mm_real *x)
 {
 	if (x) {
-		if (mm_real_is_sparse (x)) {
-			if (x->i) free (x->i);
-			if (x->p) free (x->p);
-		}
+		if (x->i) free (x->i);
+		if (x->p) free (x->p);
 		if (x->data) free (x->data);
 		free (x);
 	}
@@ -163,26 +155,8 @@ mm_real_realloc (mm_real *x, const int nnz)
 	return true;
 }
 
-/*** set to sparse ***/
-void
-mm_real_set_sparse (mm_real *x)
-{
-	if (mm_real_is_sparse (x)) return;
-	mm_set_sparse (&(x->typecode));
-	return;
-}
-
-/*** set to dense ***/
-void
-mm_real_set_dense (mm_real *x)
-{
-	if (mm_real_is_dense (x)) return;
-	mm_set_dense (&(x->typecode));
-	return;
-}
-
 /*** set to general ***/
-void
+static void
 mm_real_set_general (mm_real *x)
 {
 	if (!mm_real_is_symmetric (x)) return;
@@ -194,10 +168,9 @@ mm_real_set_general (mm_real *x)
 /*** set to symmetric
  * by default, assume symmetric upper
  * i.e., x->symm is set to MM_SYMMETRIC | MM_UPPER ***/
-void
+static void
 mm_real_set_symmetric (mm_real *x)
 {
-	if (x->m != x->n) error_and_exit ("mm_real_set_symmetric", "symmetric matrix must be square.", __FILE__, __LINE__);
 	if (mm_real_is_symmetric (x)) return;
 	mm_set_symmetric (&(x->typecode));
 	x->symm = MM_SYMMETRIC | MM_UPPER;	// by default, assume symmetric upper
@@ -205,22 +178,18 @@ mm_real_set_symmetric (mm_real *x)
 }
 
 /*** set to symmetric upper ***/
-void
+static void
 mm_real_set_upper (mm_real *x)
 {
-	if (x->m != x->n) error_and_exit ("mm_real_set_upper", "symmetric matrix must be square.", __FILE__, __LINE__);
-	if (!mm_real_is_symmetric (x)) error_and_exit ("mm_real_set_upper", "matrix must be symmetric.", __FILE__, __LINE__);
 	if (mm_real_is_upper (x)) return;
 	x->symm = MM_SYMMETRIC | MM_UPPER;
 	return;
 }
 
 /*** set to symmetric lower ***/
-void
+static void
 mm_real_set_lower (mm_real *x)
 {
-	if (x->m != x->n) error_and_exit ("mm_real_set_lower", "symmetric matrix must be square.", __FILE__, __LINE__);
-	if (!mm_real_is_symmetric (x)) error_and_exit ("mm_real_set_lower", "matrix must be symmetric.", __FILE__, __LINE__);
 	if (mm_real_is_lower (x)) return;
 	x->symm = MM_SYMMETRIC | MM_LOWER;
 	return;
@@ -231,11 +200,18 @@ static mm_sparse *
 mm_real_copy_sparse (const mm_sparse *src)
 {
 	int			k;
+	int			n = src->n;
+	int			nnz = src->nnz;
 	mm_sparse	*dest = mm_real_new (MM_REAL_SPARSE, src->symm, src->m, src->n, src->nnz);
 
-	for (k = 0; k < src->nnz; k++) dest->i[k] = src->i[k];
-	for (k = 0; k <= src->n; k++) dest->p[k] = src->p[k];
-	dcopy_ (&src->nnz, src->data, &ione, dest->data, &ione);
+	int			*si = src->i;
+	int			*sp = src->p;
+	int			*di = dest->i;
+	int			*dp = dest->p;
+
+	for (k = 0; k < nnz; k++) di[k] = si[k];
+	for (k = 0; k <= n; k++) dp[k] = sp[k];
+	dcopy_ (&nnz, src->data, &ione, dest->data, &ione);
 
 	return dest;
 }
@@ -277,11 +253,13 @@ mm_real_set_all (mm_real *x, const double val)
 mm_dense *
 mm_real_sparse_to_dense (const mm_sparse *s)
 {
-	int			j;
+	int			j, k;
+	int			p, n;
 	mm_dense	*d;
 
-	int			*si = s->i;
-	double		*sd = s->data;
+	int			*si;
+	double		*sd;
+	int			*sp = s->p;
 	double		*dd;
 
 	if (mm_real_is_dense (s)) return mm_real_copy (s);
@@ -290,9 +268,12 @@ mm_real_sparse_to_dense (const mm_sparse *s)
 
 	dd = d->data;
 	for (j = 0; j < s->n; j++) {
-		int		k = s->p[j];
-		int		pend = s->p[j + 1];
-		for (; k < pend; k++) dd[si[k] + j * s->m] = sd[k];
+		p = sp[j];
+		n = sp[j + 1] - p;
+		si = s->i + p;
+		sd = s->data + p;
+		dd = d->data + j * s->m;
+		for (k = 0; k < n; k++) dd[si[k]] = sd[k];
 	}
 	return d;
 }
@@ -302,37 +283,64 @@ mm_real_sparse_to_dense (const mm_sparse *s)
 mm_sparse *
 mm_real_dense_to_sparse (const mm_dense *d, const double threshold)
 {
-	int			j, k;
+	int			i, j, k;
 	mm_sparse	*s;
 
 	int			*si;
+	int			*sp;
 	double		*sd;
+
+	int			dm = d->m;
+	int			dn = d->n;
 	double		*dd = d->data;
+	double		dij;
 
 	if (mm_real_is_sparse (d)) return mm_real_copy (d);
 	s = mm_real_new (MM_REAL_SPARSE, d->symm, d->m, d->n, d->nnz);
 
 	si = s->i;
+	sp = s->p;
 	sd = s->data;
 
 	k = 0;
-	for (j = 0; j < d->n; j++) {
-		int		i;
-		int		i0 = 0;
-		int		i1 = d->m;
-		if (mm_real_is_symmetric (d)) {
-			if (mm_real_is_lower (d)) i0 = j;
-			if (mm_real_is_upper (d)) i1 = j + 1;
+	if (!mm_real_is_symmetric (d)) {
+		for (j = 0; j < dn; j++) {
+			for (i = 0; i < dm; i++) {
+				dij = dd[i + j * dm];
+				if (fabs (dij) >= threshold) {
+					si[k] = i;
+					sd[k] = dij;
+					k++;
+				}
+			}
+			sp[j + 1] = k;
 		}
-		for (i = i0; i < i1; i++) {
-			double	dij = dd[i + j * d->m];
-			if (fabs (dij) >= threshold) {
-				si[k] = i;
-				sd[k] = dij;
-				k++;
+	} else {
+		if (mm_real_is_upper (d)) {
+			for (j = 0; j < dn; j++) {
+				for (i = 0; i < j + 1; i++) {
+					dij = dd[i + j * dm];
+					if (fabs (dij) >= threshold) {
+						si[k] = i;
+						sd[k] = dij;
+						k++;
+					}
+				}
+				sp[j + 1] = k;
+			}
+		} else if (mm_real_is_lower (d)) {
+			for (j = 0; j < dn; j++) {
+				for (i = j; i < dm; i++) {
+					dij = dd[i + j * dm];
+					if (fabs (dij) >= threshold) {
+						si[k] = i;
+						sd[k] = dij;
+						k++;
+					}
+				}
+				sp[j + 1] = k;
 			}
 		}
-		s->p[j + 1] = k;
 	}
 	if (s->nnz != k) mm_real_realloc (s, k);
 	return s;
@@ -342,12 +350,14 @@ mm_real_dense_to_sparse (const mm_dense *d, const double threshold)
 static int
 find_jth_row_element_of_sk (const int j, const mm_sparse *s, const int k)
 {
-	int		l = s->p[k];
-	int		pend = s->p[k + 1];
-	for (; l < pend; l++) {
-		int		si = s->i[l];
-		if (si < j) continue;
-		else if (si == j) return l;	// found
+	int		l;
+	int		p = s->p[k];
+	int		n = s->p[k + 1] - p;
+	int		*si = s->i + p;
+	for (l = 0; l < n; l++) {
+		int		sil = si[l];
+		if (sil < j) continue;
+		if (sil == j) return l + p;	// found
 		else break;
 	}
 	return -1;	// not found
@@ -361,28 +371,30 @@ mm_real_symmetric_to_general_sparse (const mm_sparse *x)
 	mm_sparse	*s;
 
 	int			*si;
+	int			*sp;
 	double		*sd;
-	int			*xi;
-	double		*xd;
+	int			xn = x->n;
+	int			*xi = x->i;
+	int			*xp = x->p;
+	double		*xd = x->data;
 
 	if (!mm_real_is_symmetric (x)) return mm_real_copy (x);
 	s = mm_real_new (MM_REAL_SPARSE, MM_REAL_GENERAL, x->m, x->n, 2 * x->nnz);
 
 	si = s->i;
+	sp = s->p;
 	sd = s->data;
-	xi = x->i;
-	xd = x->data;
 
 	m = 0;
-	for (j = 0; j < x->n; j++) {
+	for (j = 0; j < xn; j++) {
 		int		k;
-		int		pend = x->p[j + 1];
+		int		pend = xp[j + 1];
 		if (mm_real_is_upper (x)) {
-			for (k = x->p[j]; k < pend; k++) {
+			for (k = xp[j]; k < pend; k++) {
 				si[m] = xi[k];
 				sd[m++] = xd[k];
 			}
-			for (k = j + 1; k < x->n; k++) {
+			for (k = j + 1; k < xn; k++) {
 				int		l = find_jth_row_element_of_sk (j, x, k);
 				// if found
 				if (l >= 0) {
@@ -399,12 +411,12 @@ mm_real_symmetric_to_general_sparse (const mm_sparse *x)
 					sd[m++] = xd[l];
 				}
 			}
-			for (k = x->p[j]; k < pend; k++) {
+			for (k = xp[j]; k < pend; k++) {
 				si[m] = xi[k];
 				sd[m++] = xd[k];
 			}
 		}
-		s->p[j + 1] = m;
+		sp[j + 1] = m;
 	}
 	if (s->nnz != m) mm_real_realloc (s, m);
 	return s;
@@ -418,14 +430,14 @@ mm_real_symmetric_to_general_dense (const mm_dense *x)
 	mm_dense	*d = mm_real_copy (x);
 	if (!mm_real_is_symmetric (x)) return d;
 
-	for (j = 0; j < x->n; j++) {
-		if (mm_real_is_upper (x)) {
+	if (mm_real_is_upper (x)) {
+		for (j = 0; j < x->n; j++) {
 			int		i0 = j + 1;
-			int		len = x->m - i0;
-			dcopy_ (&len, x->data + j + i0 * x->m, &x->m, d->data + i0 + j * d->m, &ione);
-		} else if (j > 0 && mm_real_is_lower (x)) {
-			dcopy_ (&j, x->data + j, &x->m, d->data + j * d->m, &ione);
+			int		n = x->m - i0;
+			dcopy_ (&n, x->data + j + i0 * x->m, &x->m, d->data + i0 + j * d->m, &ione);
 		}
+	} else {
+		for (j = 0; j < x->n; j++) dcopy_ (&j, x->data + j, &x->m, d->data + j * d->m, &ione);
 	}
 	mm_real_set_general (d);
 	return d;
@@ -444,10 +456,13 @@ mm_real_seye (const int n)
 {
 	int			k;
 	mm_sparse	*s = mm_real_new (MM_REAL_SPARSE, MM_REAL_GENERAL, n, n, n);
+	int			*si = s->i;
+	int			*sp = s->p;
+	double		*sd = s->data;
 	for (k = 0; k < n; k++) {
-		s->i[k] = k;
-		s->data[k] = 1.;
-		s->p[k + 1] = k + 1;
+		si[k] = k;
+		sd[k] = 1.;
+		sp[k + 1] = k + 1;
 	}
 	return s;
 }
@@ -458,8 +473,9 @@ mm_real_deye (const int n)
 {
 	int			k;
 	mm_dense	*d = mm_real_new (MM_REAL_DENSE, MM_REAL_GENERAL, n, n, n * n);
+	double		*dd = d->data;
 	mm_real_set_all (d, 0.);
-	for (k = 0; k < n; k++) d->data[k + k * n] = 1.;
+	for (k = 0; k < n; k++) dd[k + k * n] = 1.;
 	return d;
 }
 
@@ -481,20 +497,26 @@ mm_real_vertcat_sparse (const mm_sparse *s1, const mm_sparse *s2)
 	int			nnz = s1->nnz + s2->nnz;
 	mm_sparse	*s = mm_real_new (MM_REAL_SPARSE, MM_REAL_GENERAL, m, n, nnz);
 
+	int			*si = s->i;
+	int			*sp = s->p;
+	int			*si1 = s1->i;
+	int			*sp1 = s1->p;
+	int			*si2 = s2->i;
+	int			*sp2 = s2->p;
+
 	k = 0;
 	for (j = 0; j < n; j++) {
-		int		len1 = s1->p[j + 1] - s1->p[j];
-		int		len2 = s2->p[j + 1] - s2->p[j];
+		int		n1 = sp1[j + 1] - sp1[j];
+		int		n2 = sp2[j + 1] - sp2[j];
 		int		pend;
-		dcopy_ (&len1, s1->data + s1->p[j], &ione, s->data + k, &ione);
-		pend = s1->p[j + 1];
-		for (i = s1->p[j]; i < pend; i++) s->i[k++] = s1->i[i];
-		dcopy_ (&len2, s2->data + s2->p[j], &ione, s->data + k, &ione);
-		pend = s2->p[j + 1];
-		for (i = s2->p[j]; i < pend; i++) s->i[k++] = s2->i[i] + s1->m;
-		s->p[j + 1] = k;
+		dcopy_ (&n1, s1->data + sp1[j], &ione, s->data + k, &ione);
+		pend = sp1[j + 1];
+		for (i = sp1[j]; i < pend; i++) si[k++] = si1[i];
+		dcopy_ (&n2, s2->data + sp2[j], &ione, s->data + k, &ione);
+		pend = sp2[j + 1];
+		for (i = sp2[j]; i < pend; i++) si[k++] = si2[i] + s1->m;
+		sp[j + 1] = k;
 	}
-	if (s->nnz != k) mm_real_realloc (s, k);
 	return s;
 }
 
@@ -538,12 +560,21 @@ mm_real_holzcat_sparse (const mm_sparse *s1, const mm_sparse *s2)
 	int			nnz = s1->nnz + s2->nnz;
 	mm_sparse	*s = mm_real_new (MM_REAL_SPARSE, MM_REAL_GENERAL, m, n, nnz);
 
-	for (k = 0; k < s1->nnz; k++) s->i[k] = s1->i[k];
-	for (k = 0; k < s2->nnz; k++) s->i[k + s1->nnz] = s2->i[k];
-	for (j = 0; j <= s1->n; j++) s->p[j] = s1->p[j];
-	for (j = 0; j <= s2->n; j++) s->p[j + s1->n] = s2->p[j] + s1->nnz;
-	dcopy_ (&s1->nnz, s1->data, &ione, s->data, &ione);
-	dcopy_ (&s2->nnz, s2->data, &ione, s->data + s1->nnz, &ione);
+	int			*si = s->i;
+	int			*sp = s->p;
+	int			nnz1 = s1->nnz;
+	int			*si1 = s1->i;
+	int			*sp1 = s1->p;
+	int			nnz2 = s2->nnz;
+	int			*si2 = s2->i;
+	int			*sp2 = s2->p;
+
+	for (k = 0; k < nnz1; k++) si[k] = si1[k];
+	for (k = 0; k < nnz2; k++) si[k + nnz1] = si2[k];
+	for (j = 0; j <= s1->n; j++) sp[j] = sp1[j];
+	for (j = 0; j <= s2->n; j++) sp[j + s1->n] = sp2[j] + nnz1;
+	dcopy_ (&nnz1, s1->data, &ione, s->data, &ione);
+	dcopy_ (&nnz2, s2->data, &ione, s->data + nnz1, &ione);
 
 	return s;
 }
@@ -552,10 +583,7 @@ mm_real_holzcat_sparse (const mm_sparse *s1, const mm_sparse *s2)
 static mm_dense *
 mm_real_holzcat_dense (const mm_dense *d1, const mm_dense *d2)
 {
-	int			m = d1->m;
-	int			n = d1->n + d2->n;
-	int			nnz = d1->nnz + d2->nnz;
-	mm_dense	*d = mm_real_new (MM_REAL_DENSE, MM_REAL_GENERAL, m, n, nnz);
+	mm_dense	*d = mm_real_new (MM_REAL_DENSE, MM_REAL_GENERAL, d1->m, d1->n + d2->n, d1->nnz + d2->nnz);
 
 	dcopy_ (&d1->nnz, d1->data, &ione, d->data, &ione);
 	dcopy_ (&d2->nnz, d2->data, &ione, d->data + d1->nnz, &ione);
@@ -581,19 +609,19 @@ void
 mm_real_xj_add_const (mm_real *x, const int j, const double alpha)
 {
 	int		k;
-	int		len;
+	int		n;
 	double	*data;
 	if (mm_real_is_symmetric (x)) error_and_exit ("mm_real_xj_add_const", "matrix must be general.", __FILE__, __LINE__);
 	if (j < 0 || x->n <= j) error_and_exit ("mm_real_xj_add_const", "index out of range.", __FILE__, __LINE__);
 
 	if (mm_real_is_sparse (x)) {
-		len = x->p[j + 1] - x->p[j];
+		n = x->p[j + 1] - x->p[j];
 		data = x->data + x->p[j];
 	} else {
-		len = x->m;
+		n = x->m;
 		data = x->data + j * x->m;
 	}
-	for (k = 0; k < len; k++) data[k] += alpha;
+	for (k = 0; k < n; k++) data[k] += alpha;
 
 	return;
 }
@@ -602,19 +630,20 @@ mm_real_xj_add_const (mm_real *x, const int j, const double alpha)
 void
 mm_real_xj_scale (mm_real *x, const int j, const double alpha)
 {
-	int		len;
+	int		n;
 	double	*data;
 	if (mm_real_is_symmetric (x)) error_and_exit ("mm_real_xj_scale", "matrix must be general.", __FILE__, __LINE__);
 	if (j < 0 || x->n <= j) error_and_exit ("mm_real_xj_scale", "index out of range.", __FILE__, __LINE__);
 
 	if (mm_real_is_sparse (x)) {
-		len = x->p[j + 1] - x->p[j];
-		data = x->data + x->p[j];
+		int		p = x->p[j];
+		n = x->p[j + 1] - p;
+		data = x->data + p;
 	} else {
-		len = x->m;
+		n = x->m;
 		data = x->data + j * x->m;
 	}
-	dscal_ (&len, &alpha, data, &ione);
+	dscal_ (&n, &alpha, data, &ione);
 
 	return;
 }
@@ -623,9 +652,10 @@ mm_real_xj_scale (mm_real *x, const int j, const double alpha)
 static double
 mm_real_sj_asum (const mm_sparse *s, const int j)
 {
-	int		size = s->p[j + 1] - s->p[j];
+	int		p = s->p[j];
+	int		n = s->p[j + 1] - p;
 	double	*sd = s->data;
-	double	asum = dasum_ (&size, sd + s->p[j], &ione);
+	double	asum = dasum_ (&n, sd + p, &ione);
 	if (mm_real_is_symmetric (s)) {
 		int		k;
 		int		k0;
@@ -653,17 +683,17 @@ mm_real_dj_asum (const mm_dense *d, const int j)
 	double	val = 0.;
 	if (!mm_real_is_symmetric (d)) val = dasum_ (&d->m, d->data + j * d->m, &ione);
 	else {
-		int		len;
+		int		n;
 		if (mm_real_is_upper (d)) {
-			len = j;
-			val = dasum_ (&len, d->data + j * d->m, &ione);
-			len = d->m - j;
-			val += dasum_ (&len, d->data + j * d->m + j, &d->m);
+			n = j;
+			val = dasum_ (&n, d->data + j * d->m, &ione);
+			n = d->m - j;
+			val += dasum_ (&n, d->data + j * d->m + j, &d->m);
 		} else if (mm_real_is_lower (d)) {
-			len = d->m - j;
-			val = dasum_ (&len, d->data + j * d->m + j, &ione);
-			len = j;
-			val += dasum_ (&len, d->data + j, &d->m);
+			n = d->m - j;
+			val = dasum_ (&n, d->data + j * d->m + j, &ione);
+			n = j;
+			val += dasum_ (&n, d->data + j, &d->m);
 		}
 	}
 	return val;
@@ -681,11 +711,12 @@ mm_real_xj_asum (const mm_real *x, const int j)
 static double
 mm_real_sj_sum (const mm_sparse *s, const int j)
 {
-	int		k = s->p[j];
-	int		pend = s->p[j + 1];
-	double	*sd = s->data;
+	int		k;
+	int		p = s->p[j];
+	int		n = s->p[j + 1] - p;
+	double	*sd = s->data + p;
 	double	sum = 0.;
-	for (; k < pend; k++) sum += sd[k];
+	for (k = 0; k < n; k++) sum += sd[k];
 	if (mm_real_is_symmetric (s)) {
 		int		k0;
 		int		k1;
@@ -696,6 +727,7 @@ mm_real_sj_sum (const mm_sparse *s, const int j)
 			k0 = 0;
 			k1 = j;
 		}
+		sd = s->data;
 		for (k = k0; k < k1; k++) {
 			int		l = find_jth_row_element_of_sk (j, s, k);
 			// if found
@@ -710,21 +742,27 @@ static double
 mm_real_dj_sum (const mm_dense *d, const int j)
 {
 	int		k;
+	double	*dd;
 	double	sum = 0.;
-	if (!mm_real_is_symmetric (d))
-		for (k = 0; k < d->m; k++) sum += d->data[k + j * d->m];
-	else {
-		int		len;
+	if (!mm_real_is_symmetric (d)) {
+		dd = d->data + j * d->m;
+		for (k = 0; k < d->m; k++) sum += dd[k];
+	} else {
+		int		n;
 		if (mm_real_is_upper (d)) {
-			len = j;
-			for (k = 0; k < len; k++) sum += d->data[k + j * d->m];
-			len = d->m - j;
-			for (k = 0; k < len; k++) sum += d->data[k * d->m + j * d->m + j];
+			n = j;
+			dd = d->data + j * d->m;
+			for (k = 0; k < n; k++) sum += dd[k];
+			n = d->m - j;
+			dd = d->data + j + j * d->m;
+			for (k = 0; k < n; k++) sum += dd[k * d->m];
 		} else if (mm_real_is_lower (d)) {
-			len = d->m - j;
-			for (k = 0; k < len; k++) sum += d->data[k + j * d->m + j];
-			len = j;
-			for (k = 0; k < len; k++) sum += d->data[k * d->m + j];
+			n = d->m - j;
+			dd = d->data + j + j * d->m;
+			for (k = 0; k < n; k++) sum += dd[k];
+			n = j;
+			dd = d->data + j;
+			for (k = 0; k < n; k++) sum += dd[k * d->m];
 		}
 	}
 	return sum;
@@ -742,9 +780,10 @@ mm_real_xj_sum (const mm_real *x, const int j)
 static double
 mm_real_sj_ssq (const mm_sparse *s, const int j)
 {
-	int		size = s->p[j + 1] - s->p[j];
+	int		p = s->p[j];
+	int		n = s->p[j + 1] - p;
 	double	*sd = s->data;
-	double	ssq = ddot_ (&size, sd + s->p[j], &ione, sd + s->p[j], &ione);
+	double	ssq = ddot_ (&n, sd + p, &ione, sd + p, &ione);
 	if (mm_real_is_symmetric (s)) {
 		int		k;
 		int		k0;
@@ -772,18 +811,18 @@ mm_real_dj_ssq (const mm_dense *d, const int j)
 	double	ssq;
 	if (!mm_real_is_symmetric (d)) ssq = ddot_ (&d->m, d->data + j * d->m, &ione, d->data + j * d->m, &ione);
 	else {
-		int		len;
+		int		n;
 		ssq = 0.;
 		if (mm_real_is_upper (d)) {
-			len = j;
-			ssq = ddot_ (&len, d->data + j * d->m, &ione, d->data + j * d->m, &ione);
-			len = d->m - j;
-			ssq += ddot_ (&len, d->data + j * d->m + j, &d->m, d->data + j * d->m + j, &d->m);
+			n = j;
+			ssq = ddot_ (&n, d->data + j * d->m, &ione, d->data + j * d->m, &ione);
+			n = d->m - j;
+			ssq += ddot_ (&n, d->data + j * d->m + j, &d->m, d->data + j * d->m + j, &d->m);
 		} else if (mm_real_is_lower (d)) {
-			len = d->m - j;
-			ssq = ddot_ (&len, d->data + j * d->m + j, &ione, d->data + j * d->m + j, &ione);
-			len = j;
-			ssq += ddot_ (&len, d->data + j, &d->m, d->data + j, &d->m);
+			n = d->m - j;
+			ssq = ddot_ (&n, d->data + j * d->m + j, &ione, d->data + j * d->m + j, &ione);
+			n = j;
+			ssq += ddot_ (&n, d->data + j, &d->m, d->data + j, &d->m);
 		}
 	}
 	return ssq;
@@ -807,77 +846,163 @@ mm_real_xj_nrm2 (const mm_real *x, const int j)
 	return sqrt (ssq);
 }
 
-/* alpha * s * y, where s is sparse matrix and y is dense vector */
-static mm_dense *
-mm_real_s_dot_y (bool trans, const double alpha, const mm_sparse *s, const mm_dense *y)
+/* z = alpha * s * y + beta * z, where s is sparse matrix and y is dense vector */
+static void
+mm_real_s_dot_y (bool trans, const double alpha, const mm_sparse *s, const mm_dense *y, const double beta, mm_dense *z)
 {
-	int			j;
-	int			m;
-	mm_dense	*c;
+	int			j, k;
+	int			p, n;
 
-	double		*sd = s->data;
+	int			*si;
+	double		*sd;
+	int			*sp = s->p;
 	double		*yd = y->data;
-	double		*cd;
+	double		*zd = z->data;
 
-	m = (trans) ? s->n : s->m;
+	if (fabs (beta) > 0.) mm_real_xj_scale (z, 0, beta);
+	else mm_real_set_all (z, 0.);
 
-	c = mm_real_new (MM_REAL_DENSE, MM_REAL_GENERAL, m, 1, m);
-	mm_real_set_all (c, 0.);
-
-	cd = c->data;
-	for (j = 0; j < s->n; j++) {
-		int		k = s->p[j];
-		int		pend = s->p[j + 1];
-		for (; k < pend; k++) {
-			int		si;
-			int		sj;
-			if (trans) {
-				si = j;
-				sj = s->i[k];
-			} else {
-				si = s->i[k];
-				sj = j;
+	if (trans) {
+		if (!mm_real_is_symmetric (s)) {
+			for (j = 0; j < s->n; j++) {
+				p = sp[j];
+				n = sp[j + 1] - p;
+				if (n <= 0) continue;
+				si = s->i + p;
+				sd = s->data + p;
+				for (k = 0; k < n; k++) zd[j] += alpha * sd[k] * yd[si[k]];
 			}
-			cd[si] += alpha * sd[k] * yd[sj];
-			if (mm_real_is_symmetric (s) && j != s->i[k]) cd[sj] += alpha * sd[k] * yd[si];
-		}		
-	}
-	return c;
+		} else if (mm_real_is_upper (s)) {
+			int		sik;
+			double	alpha_sdk;
+			for (j = 0; j < s->n; j++) {
+				p = sp[j];
+				n = sp[j + 1] - p;
+				si = s->i + p;
+				sd = s->data + p;
+				for (k = 0; k < n - 1; k++) {
+					sik = si[k];
+					alpha_sdk = alpha * sd[k];
+					zd[j] += alpha_sdk * yd[sik];
+					zd[sik] += alpha_sdk * yd[j];
+				}
+				if (n <= 0) continue;
+				sik = si[n - 1];
+				alpha_sdk = alpha * sd[n - 1];
+				zd[j] += alpha_sdk * yd[sik];
+				// omit if diagonal element
+				if (j != sik) zd[sik] += alpha_sdk * yd[j];
+			}
+		} else {
+			int		sik;
+			double	alpha_sdk;
+			for (j = 0; j < s->n; j++) {
+				p = sp[j];
+				n = sp[j + 1] - p;
+				if (n <= 0) continue;
+				si = s->i + p;
+				sd = s->data + p;
+				sik = si[0];
+				alpha_sdk = alpha * sd[0];
+				zd[j] += alpha_sdk * yd[sik];
+				// omit if diagonal element
+				if (j != sik) zd[sik] += alpha_sdk * yd[j];
+				for (k = 1; k < n; k++) {
+					sik = si[k];
+					alpha_sdk = alpha * sd[k];
+					zd[j] += alpha_sdk * yd[sik];
+					zd[sik] += alpha_sdk * yd[j];
+				}
+			}
+		}
+	} else {
+		if (!mm_real_is_symmetric (s)) {
+			for (j = 0; j < s->n; j++) {
+				p = sp[j];
+				n = sp[j + 1] - p;
+				if (n <= 0) continue;
+				si = s->i + p;
+				sd = s->data + p;
+				for (k = 0; k < n; k++) zd[si[k]] += alpha * sd[k] * yd[j];
+			}
+		} else if (mm_real_is_upper (s)) {
+			int		sik;
+			double	alpha_sdk;
+			for (j = 0; j < s->n; j++) {
+				p = sp[j];
+				n = sp[j + 1] - p;
+				si = s->i + p;
+				sd = s->data + p;
+				for (k = 0; k < n - 1; k++) {
+					sik = si[k];
+					alpha_sdk = alpha * sd[k];
+					zd[sik] += alpha_sdk * yd[j];
+					zd[j] += alpha_sdk * yd[sik];
+				}
+				if (n <= 0) continue;
+				sik = si[n - 1];
+				alpha_sdk = alpha * sd[n - 1];
+				zd[sik] += alpha_sdk * yd[j];
+				// omit if diagonal element
+				if (j != sik) zd[j] += alpha_sdk * yd[sik];
+			}
+		} else {
+			int		sik;
+			double	alpha_sdk;
+			for (j = 0; j < s->n; j++) {
+				p = sp[j];
+				n = sp[j + 1] - p;
+				if (n <= 0) continue;
+				si = s->i + p;
+				sd = s->data + p;
+				sik = si[0];
+				alpha_sdk = alpha * sd[0];
+				zd[sik] += alpha_sdk * yd[j];
+				// omit if diagonal element
+				if (j != sik) zd[j] += alpha_sdk * yd[sik];
+				for (k = 1; k < n; k++) {
+					sik = si[k];
+					alpha_sdk = alpha * sd[k];
+					zd[sik] += alpha_sdk * yd[j];
+					zd[j] += alpha_sdk * yd[sik];
+				}
+			}
+		}
+	}	
+	return;
 }
 
-/* alpha * d * y, where d is dense matrix and y is dense vector */
-static mm_dense *
-mm_real_d_dot_y (bool trans, const double alpha, const mm_dense *d, const mm_dense *y)
+/* z = alpha * d * y + beta * z, where d is dense matrix and y is dense vector */
+static void
+mm_real_d_dot_y (bool trans, const double alpha, const mm_dense *d, const mm_dense *y, const double beta, mm_dense *z)
 {
-	int			m;
-	mm_dense	*c;
-	m = (trans) ? d->n : d->m;
-
-	c = mm_real_new (MM_REAL_DENSE, MM_REAL_GENERAL, m, 1, m);
-	mm_real_set_all (c, 0.);
-
 	if (!mm_real_is_symmetric (d)) {
-		// c = alpha*d*y + 0*c
-		dgemv_ ((trans) ? "T" : "N", &d->m, &d->n, &alpha, d->data, &d->m, y->data, &ione, &dzero, c->data, &ione);
+		// z = alpha * d * y + beta * z
+		dgemv_ ((trans) ? "T" : "N", &d->m, &d->n, &alpha, d->data, &d->m, y->data, &ione, &beta, z->data, &ione);
 	} else {
 		char	uplo = (mm_real_is_upper (d)) ? 'U' : 'L';
-		// c = alpha*d*y + 0*c
-		dsymv_ (&uplo, &d->m, &alpha, d->data, &d->m, y->data, &ione, &dzero, c->data, &ione);
+		// z = alpha * d * y + beta * z
+		dsymv_ (&uplo, &d->m, &alpha, d->data, &d->m, y->data, &ione, &beta, z->data, &ione);
 	}
-	return c;
+	return;
 }
 
 /*** alpha * x * y, where x is sparse/dense matrix and y is dense vector ***/
-mm_dense *
-mm_real_x_dot_y (bool trans, const double alpha, const mm_real *x, const mm_dense *y)
+void
+mm_real_x_dot_y (bool trans, const double alpha, const mm_real *x, const mm_dense *y, const double beta, mm_dense *z)
 {
 	if (!mm_real_is_dense (y)) error_and_exit ("mm_real_x_dot_y", "y must be dense.", __FILE__, __LINE__);
 	if (mm_real_is_symmetric (y)) error_and_exit ("mm_real_x_dot_y", "y must be general.", __FILE__, __LINE__);
 	if (y->n != 1) error_and_exit ("mm_real_x_dot_y", "y must be vector.", __FILE__, __LINE__);
 	if ((trans && x->m != y->m) || (!trans && x->n != y->m))
-		error_and_exit ("mm_real_x_dot_y", "vector and matrix dimensions do not match.", __FILE__, __LINE__);
+		error_and_exit ("mm_real_x_dot_y", "dimensions of x and y do not match.", __FILE__, __LINE__);
+	if (!mm_real_is_dense (z)) error_and_exit ("mm_real_x_dot_y", "z must be dense.", __FILE__, __LINE__);
+	if (mm_real_is_symmetric (z)) error_and_exit ("mm_real_x_dot_y", "z must be general.", __FILE__, __LINE__);
+	if (z->n != 1) error_and_exit ("mm_real_x_dot_y", "z must be vector.", __FILE__, __LINE__);
+	if ((trans && x->n != z->m) || (!trans && x->m != z->m))
+		error_and_exit ("mm_real_x_dot_y", "dimensions of x and z do not match.", __FILE__, __LINE__);
 
-	return (mm_real_is_sparse (x)) ? mm_real_s_dot_y (trans, alpha, x, y) : mm_real_d_dot_y (trans, alpha, x, y);
+	return (mm_real_is_sparse (x)) ? mm_real_s_dot_y (trans, alpha, x, y, beta, z) : mm_real_d_dot_y (trans, alpha, x, y, beta, z);
 }
 
 /* s(:,j)' * y */
@@ -886,13 +1011,37 @@ mm_real_sj_trans_dot_y (const mm_sparse *s, const int j, const mm_dense *y)
 {
 	double	val = 0;
 
-	int		*si = s->i;
-	double	*sd = s->data;
+	int		*si;
+	double	*sd;
 	double	*yd = y->data;
 
-	int		k = s->p[j];
-	int		pend = s->p[j + 1];
-	for (; k < pend; k++) val += sd[k] * yd[si[k]];
+	int		k;
+	int		p = s->p[j];
+	int		n = s->p[j + 1] - p;
+	int		mod = n % 4;
+
+	// unrolling
+	si = s->i + p;
+	sd = s->data + p;
+	if (mod == 1) {
+		val += sd[0] * yd[si[0]];
+		si++;
+		sd++;
+	} else if (mod == 2) {
+		val += sd[0] * yd[si[0]] + sd[1] * yd[si[1]];
+		si += 2;
+		sd += 2;
+	} else if (mod == 3) {
+		val += sd[0] * yd[si[0]] + sd[1] * yd[si[1]] + sd[2] * yd[si[2]];
+		si += 3;
+		sd += 3;
+	}
+	for (k = mod; k < n; k += 4) {
+		val += sd[0] * yd[si[0]] + sd[1] * yd[si[1]] + sd[2] * yd[si[2]] + sd[3] * yd[si[3]];
+		si += 4;
+		sd += 4;
+	}
+
 	if (mm_real_is_symmetric (s)) {
 		int		k0;
 		int		k1;
@@ -903,12 +1052,14 @@ mm_real_sj_trans_dot_y (const mm_sparse *s, const int j, const mm_dense *y)
 			k0 = 0;
 			k1 = j;
 		}
+		sd = s->data;
 		for (k = k0; k < k1; k++) {
 			int		l = find_jth_row_element_of_sk (j, s, k);
 			// if found
 			if (l >= 0) val += sd[l] * yd[k];
 		}
 	}
+
 	return val;
 }
 
@@ -919,17 +1070,17 @@ mm_real_dj_trans_dot_y (const mm_dense *d, const int j, const mm_dense *y)
 	double	val = 0.;
 	if (!mm_real_is_symmetric (d)) val = ddot_ (&d->m, d->data + j * d->m, &ione, y->data, &ione);
 	else {
-		int		len;
+		int		n;
 		if (mm_real_is_upper (d)) {
-			len = j;
-			val = ddot_ (&len, d->data + j * d->m, &ione, y->data, &ione);
-			len = d->m - j;
-			val += ddot_ (&len, d->data + j * d->m + j, &d->m, y->data + j, &ione);
+			n = j;
+			val = ddot_ (&n, d->data + j * d->m, &ione, y->data, &ione);
+			n = d->m - j;
+			val += ddot_ (&n, d->data + j * d->m + j, &d->m, y->data + j, &ione);
 		} else if (mm_real_is_lower (d)) {
-			len = d->m - j;
-			val = ddot_ (&len, d->data + j * d->m + j, &ione, y->data + j, &ione);
-			len = j;
-			val += ddot_ (&len, d->data + j, &d->m, y->data, &ione);
+			n = d->m - j;
+			val = ddot_ (&n, d->data + j * d->m + j, &ione, y->data + j, &ione);
+			n = j;
+			val += ddot_ (&n, d->data + j, &d->m, y->data, &ione);
 		}
 	}
 	return val;
@@ -952,13 +1103,16 @@ mm_real_xj_trans_dot_y (const mm_real *x, const int j, const mm_dense *y)
 static void
 mm_real_asjpy (const double alpha, const mm_sparse *s, const int j, mm_dense *y)
 {
-	int		*si = s->i;
-	double	*sd = s->data;
+	int		*si;
+	double	*sd;
 	double	*yd = y->data;
 
-	int		k = s->p[j];
-	int		pend = s->p[j + 1];
-	for (; k < pend; k++) yd[si[k]] += alpha * sd[k];
+	int		k;
+	int		p = s->p[j];
+	int		n = s->p[j + 1] - p;
+	si = s->i + p;
+	sd = s->data + p;
+	for (k = 0; k < n; k++) yd[si[k]] += alpha * sd[k];
 	if (mm_real_is_symmetric (s)) {
 		int		k0;
 		int		k1;
@@ -969,6 +1123,7 @@ mm_real_asjpy (const double alpha, const mm_sparse *s, const int j, mm_dense *y)
 			k0 = 0;
 			k1 = j;
 		}
+		sd = s->data;
 		for (k = k0; k < k1; k++) {
 			int		l = find_jth_row_element_of_sk (j, s, k);
 			// if found
@@ -984,17 +1139,17 @@ mm_real_adjpy (const double alpha, const mm_dense *d, const int j, mm_dense *y)
 {
 	if (!mm_real_is_symmetric (d)) daxpy_ (&d->m, &alpha, d->data + j * d->m, &ione, y->data, &ione);
 	else {
-		int		len;
+		int		n;
 		if (mm_real_is_upper (d)) {
-			len = j;
-			daxpy_ (&len, &alpha, d->data + j * d->m, &ione, y->data, &ione);
-			len = d->m - j;
-			daxpy_ (&len, &alpha, d->data + j * d->m + j, &d->m, y->data + j, &ione);
+			n = j;
+			daxpy_ (&n, &alpha, d->data + j * d->m, &ione, y->data, &ione);
+			n = d->m - j;
+			daxpy_ (&n, &alpha, d->data + j * d->m + j, &d->m, y->data + j, &ione);
 		} else if (mm_real_is_lower (d)) {
-			len = d->m - j;
-			daxpy_ (&len, &alpha, d->data + j * d->m + j, &ione, y->data + j, &ione);
-			len = j;
-			daxpy_ (&len, &alpha, d->data + j, &d->m, y->data, &ione);
+			n = d->m - j;
+			daxpy_ (&n, &alpha, d->data + j * d->m + j, &ione, y->data + j, &ione);
+			n = j;
+			daxpy_ (&n, &alpha, d->data + j, &d->m, y->data, &ione);
 		}
 	}
 	return;
@@ -1017,13 +1172,16 @@ mm_real_axjpy (const double alpha, const mm_real *x, const int j, mm_dense *y)
 static void
 mm_real_asjpy_atomic (const double alpha, const mm_sparse *s, const int j, mm_dense *y)
 {
-	int		*si = s->i;
-	double	*sd = s->data;
+	int		*si;
+	double	*sd;
 	double	*yd = y->data;
 
-	int		k = s->p[j];
-	int		pend = s->p[j + 1];
-	for (; k < pend; k++) atomic_add (yd + si[k], alpha * sd[k]);
+	int		k;
+	int		p = s->p[j];
+	int		n = s->p[j + 1] - p;
+	si = s->i + p;
+	sd = s->data + p;
+	for (k = 0; k < n; k++) atomic_add (yd + si[k], alpha * sd[k]);
 	if (mm_real_is_symmetric (s)) {
 		int		k0;
 		int		k1;
@@ -1034,6 +1192,7 @@ mm_real_asjpy_atomic (const double alpha, const mm_sparse *s, const int j, mm_de
 			k0 = 0;
 			k1 = j;
 		}
+		sd = s->data;
 		for (k = k0; k < k1; k++) {
 			int		l = find_jth_row_element_of_sk (j, s, k);
 			// if found
@@ -1047,24 +1206,29 @@ mm_real_asjpy_atomic (const double alpha, const mm_sparse *s, const int j, mm_de
 static void
 mm_real_adjpy_atomic (const double alpha, const mm_dense *d, const int j, mm_dense *y)
 {
-	double	*dd = d->data;
+	double	*dd;
 	double	*yd = y->data;
 
 	int		k;
 	if (!mm_real_is_symmetric (d)) {
-		for (k = 0; k < d->m; k++) atomic_add (yd + k, alpha * dd[j * d->m + k]);
+		dd = d->data + j * d->m;
+		for (k = 0; k < d->m; k++) atomic_add (yd + k, alpha * dd[k]);
 	} else {
-		int		len;
+		int		n;
 		if (mm_real_is_upper (d)) {
-			len = j;
-			for (k = 0; k < len; k++) atomic_add (yd + k, alpha * dd[j * d->m + k]);
-			len = d->m - j;
-			for (k = 0; k < len; k++) atomic_add (yd + j + k, alpha * dd[j * d->m + j + k * d->m]);
+			n = j;
+			dd = d->data + j * d->m;
+			for (k = 0; k < n; k++) atomic_add (yd + k, alpha * dd[k]);
+			n = d->m - j;
+			dd = d->data + j * d->m + j;
+			for (k = 0; k < n; k++) atomic_add (yd + j + k, alpha * dd[k * d->m]);
 		} else if (mm_real_is_lower (d)) {
-			len = d->m - j;
-			for (k = 0; k < len; k++) atomic_add (yd + j + k, alpha * dd[j * d->m + j + k]);
-			len = j;
-			for (k = 0; k < len; k++) atomic_add (yd + k, alpha * dd[j + k * d->m]);
+			n = d->m - j;
+			dd = d->data + j * d->m + j;
+			for (k = 0; k < n; k++) atomic_add (yd + j + k, alpha * dd[k]);
+			n = j;
+			dd = d->data + j;
+			for (k = 0; k < n; k++) atomic_add (yd + k, alpha * dd[k * d->m]);
 		}
 	}
 	return;
@@ -1103,7 +1267,7 @@ mm_real_fread_sparse (FILE *fp, MM_typecode typecode)
 	}
 
 	l = 0;
-	for (k = 0; k < s->nnz; k++) {
+	for (k = 0; k < nnz; k++) {
 		s->i[k]--;	// fortran -> c
 		while (l < j[k]) s->p[l++] = k;
 	}
@@ -1111,7 +1275,7 @@ mm_real_fread_sparse (FILE *fp, MM_typecode typecode)
 
 	if (mm_is_symmetric (typecode)) {
 		mm_real_set_symmetric (s);
-		for (k = 0; k < s->nnz; k++) {
+		for (k = 0; k < nnz; k++) {
 			if (s->i[k] == j[k] - 1) continue;
 			(s->i[k] < j[k] - 1) ? mm_real_set_upper (s) : mm_real_set_lower (s);
 			break;
@@ -1156,6 +1320,7 @@ mm_real_fread (FILE *fp)
 		error_and_exit ("mm_real_fread", msg, __FILE__, __LINE__);
 	}
 	x = (mm_is_sparse (typecode)) ? mm_real_fread_sparse (fp, typecode) : mm_real_fread_dense (fp, typecode);
+	if (!x) error_and_exit ("mm_real_fread", "failed to read mm_real.", __FILE__, __LINE__);
 	if (mm_real_is_symmetric (x) && x->m != x->n) error_and_exit ("mm_real_fread", "symmetric matrix must be square.", __FILE__, __LINE__);
 	return x;
 }
